from typing import Any, Iterable

import numpy as np
from numpy import ndarray
from sectionproperties.analysis.section import Section

from polymesh.config import __hasplotly__, __hasmatplotlib__

if __hasplotly__:
    from dewloosh.plotly import plot_lines_3d

from .mesh import FemMesh
from .cells import B2, B3


class LineMesh(FemMesh):
    """
    A data class dedicated to 1d cells. It handles sections and other line
    related information, plotting, etc.

    See Also
    --------
    :class:`sigmaepsilon.fem.mesh.FemMesh`
    """

    _cell_classes_ = {
        2: B2,
        3: B3,
    }

    def plot(self, *args, scalars=None, backend="plotly", scalar_labels=None, **kwargs):
        """
        Plots the line elements using one of the supported backends.

        Parameters
        ----------
        scalars : numpy.ndarray, Optional
            Data to plot. Default is None.

        backend : str, Optional
            The backend to use for plotting. Available options are 'plotly' and 'vtk'.
            Default is 'plotly'.

        scalar_labels : Iterable, Optional
            Labels of the scalars in 'scalars'. Only if Plotly is selected as the backend.
            Defaeult is None.

        Returns
        -------
        Any
            A PyVista or a Plotly object.

        """
        if backend == "vtk":
            return self.pvplot(
                *args, scalars=scalars, scalar_labels=scalar_labels, **kwargs
            )
        elif backend == "plotly":
            assert __hasplotly__
            coords = self.coords()
            topo = self.topology()
            return plot_lines_3d(coords, topo)
        elif backend == "mpl":
            raise NotImplementedError
            assert __hasmatplotlib__
        else:
            msg = "No implementation for backend '{}'".format(backend)
            raise NotImplementedError(msg)

    def plot_dof_solution(
        self,
        *args,
        backend: str = "plotly",
        case: int = 0,
        component: int = 0,
        labels: Iterable = None,
        **kwargs
    ) -> Any:
        """
        Plots degrees of freedom solution using 'vtk' or 'plotly'.

        Parameters
        ----------
        scalars : numpy.ndarray, Optional
            Data to plot. Default is None.
        case : int, Optional
            The index of the load case. Default is 0.
        component : int, Optional
            The index of the DOF component. Default is 0.
        labels : Iterable, Optional
            Labels of the DOFs. Only if Plotly is selected as the backend.
            Defaeult is None.
        **kwargs : dict, Optional
            Keyqord arguments forwarded to :func:`pvplot`.

        Returns
        -------
        Any
            A figure object or None, depending on the selected backend.
        """
        if backend == "vtk":
            scalars = self.nodal_dof_solution()[:, component, case]
            return self.pvplot(*args, scalars=scalars, **kwargs)
        elif backend == "plotly":
            scalar_labels = labels if labels is not None else ["U", "V", "W"]
            coords = self.coords()
            topo = self.topology()
            dofsol = self.nodal_dof_solution()[:, :3, case]
            return plot_lines_3d(
                coords, topo, scalars=dofsol, scalar_labels=scalar_labels
            )
        else:
            msg = "No implementation for backend '{}'".format(backend)
            raise NotImplementedError(msg)

    def _init_config_(self):
        super()._init_config_()
        key = self.__class__._pv_config_key_
        self.config[key]["color"] = "k"
        self.config[key]["line_width"] = 10
        self.config[key]["render_lines_as_tubes"] = True


class BernoulliFrame(LineMesh):
    """
    A subclass of :class:`LineMesh` to handle input and output
    of 1d meshes.

    Note
    ----
    This in experimental stage.

    """

    NDOFN = 6

    _cell_classes_ = {
        2: B2,
        3: B3,
    }



from typing import Union

from numba import njit, prange
import numpy as np
from numpy import ndarray
from numba import types as nbtypes
from numba.typed import Dict as nbDict

from neumann.linalg.sparse import csr_matrix

__cache = True

nbint64 = nbtypes.int64
nbint64A = nbint64[:]
nbfloat64A = nbtypes.float64[:]


@njit(nogil=True, cache=__cache)
def get_filter_factors(
    centers: ndarray, neighbours: Union[nbDict, ndarray], r_max: float
):
    nE = len(centers)
    res = nbDict.empty(
        key_type=nbint64,
        value_type=nbfloat64A,
    )
    for iE in range(nE):
        res[iE] = r_max - norms(centers[neighbours[iE]] - centers[iE])
    return res


@njit(nogil=True, cache=__cache)
def get_filter_factors_csr(centers: ndarray, neighbours: csr_matrix, r_max: float):
    nE = len(centers)
    res = nbDict.empty(
        key_type=nbint64,
        value_type=nbfloat64A,
    )
    for iE in range(nE):
        ni, _ = neighbours.row(iE)
        res[iE] = r_max - norms(centers[ni] - centers[iE])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def norms(a: ndarray):
    nI = len(a)
    res = np.zeros(nI, dtype=a.dtype)
    for iI in prange(len(a)):
        res[iI] = np.dot(a[iI], a[iI])
    return np.sqrt(res)


@njit(nogil=True, parallel=True, cache=__cache)
def norm(a: ndarray):
    return np.sqrt(np.dot(a, a))


@njit(nogil=True, parallel=True, cache=__cache)
def weighted_stiffness_bulk(K: ndarray, weights: ndarray):
    nE = len(weights)
    res = np.zeros_like(K, dtype=K.dtype)
    for i in prange(nE):
        res[i, :] = weights[i] * K[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def weighted_stiffness_flat(K: ndarray, weights: ndarray, kranges: ndarray):
    nE = len(weights)
    res = np.zeros_like(K, dtype=K.dtype)
    for i in prange(nE):
        _r, r_ = kranges[i], kranges[i + 1]
        res[_r:r_] = weights[i] * K[_r:r_]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def weighted_stiffness_1d_regular(data: ndarray, weights: ndarray):
    nE = len(weights)
    dsize = int(len(data) / nE)
    res = np.zeros_like(data, dtype=data.dtype)
    for i in prange(nE):
        res[i * dsize : (i + 1) * dsize] = (
            data[i * dsize : (i + 1) * dsize] * weights[i]
        )
    return res


def element_stiffness_ranges(kshape: tuple) -> ndarray:
    nE = kshape[0]
    if len(kshape) == 3 and isinstance(kshape[1], int):
        res = np.zeros(nE + 1, dtype=int)
        res[1:] = np.cumsum(np.full(nE, kshape[1] * kshape[2]))
    elif isinstance(kshape[1], ndarray):
        res = np.zeros(nE + 1, dtype=int)
        res[1:] = np.cumsum(kshape[1])
    else:
        raise ValueError(f"Invalid shape {kshape}")
    return res



import numpy as np
from numpy import ndarray
from typing import Union, Callable, Iterable

from scipy import interpolate

from neumann import atleast1d, atleastnd, ascont
from neumann.utils import to_range_1d

from ...utils.fem.cells.bernoulli import (
    shape_function_matrix_Bernoulli_bulk,
    body_load_vector_Bernoulli,
    global_shape_function_derivatives_Bernoulli_bulk as gdshpB,
    lumped_mass_matrices_direct_Bernoulli as dlump,
)

from ...utils.fem.postproc import (
    approx_element_solution_bulk,
    calculate_internal_forces_bulk,
)
from ..material.beam import (
    BernoulliBeam,
    _postproc_bernoulli_internal_forces_H_,
    _postproc_bernoulli_internal_forces_L_,
)

__all__ = ["BernoulliBase"]


class BernoulliBase(BernoulliBeam):
    """
    Skeleton class for 1d finite elements, whose bending behaviour
    is governed by the Bernoulli theory. The implementations covered
    by this class are independent of the number of nodes of the cells.
    Specification of an actual finite element consists of specifying class
    level attributes only.

    See Also
    --------
    :class:`~sigmaepsilon.fem.cells.bernoulli2.Bernoulli2`
    :class:`~sigmaepsilon.fem.cells.bernoulli3.Bernoulli3`
    """

    qrule: str = None
    quadrature: dict = None
    shpfnc: Callable = None
    dshpfnc: Callable = None
    dshpfnc_geom: Callable = None

    def shape_function_values(
        self,
        pcoords: Union[float, Iterable[float]],
        *_,
        rng: Iterable = None,
        lengths: Iterable[float] = None,
        **__,
    ) -> ndarray:
        """
        Evaluates the shape functions at the points specified by 'pcoords'.

        Parameters
        ----------
        pcoords : float or Iterable[float]
            Locations of the evaluation points.
        rng : Iterable, Optional
            The range in which the locations ought to be understood,
            typically [0, 1] or [-1, 1]. Default is [0, 1].
        lengths : Iterable, Optional
            The lengths of the beams in the block. Default is None.

        Returns
        -------
        numpy.ndarray
            The returned array has a shape of (nP, nNE=2, nDOF=6), where nP,
            nNE and nDOF stand for the number of evaluation points, nodes
            per element and number of degrees of freedom, respectively.
        """
        rng = np.array([-1, 1]) if rng is None else np.array(rng)
        pcoords = atleast1d(np.array(pcoords))
        pcoords = to_range_1d(pcoords, source=rng, target=[-1, 1])
        lengths = self.lengths() if lengths is None else lengths
        return self.__class__.shpfnc(pcoords, lengths).astype(float)

    def shape_function_derivatives(
        self,
        pcoords: Union[float, Iterable[float]] = None,
        *_,
        rng: Iterable = None,
        jac: ndarray = None,
        dshp: ndarray = None,
        lengths: Iterable[float] = None,
        **__,
    ) -> ndarray:
        """
        Evaluates the shape function derivatives (up to third) at the points specified
        by 'pcoords'. The function either returns the derivatives on the master or the actual
        element, depending on the inputs.

        Valid combination of inputs are:

            * 'pcoords' and optionally 'jac' : this can be used to calculate the derivatives
              in both the global ('jac' is provided) and the master frame ('jac' is not provided).

            * 'dshp' and 'jac' : this combination can only be used to return the derivatives
              wrt. to the global frame.

        Parameters
        ----------
        pcoords : float or Iterable, Optional
            Locations of the evaluation points. Default is None.
        rng : Iterable, Optional
            The range in which the locations ought to be understood,
            typically [0, 1] or [-1, 1]. Only if 'pcoords' is provided.
            Default is [0, 1].
        lengths : Iterable, Optional
            The lengths of the beams in the block. Default is None.
        jac : Iterable, Optional
            The jacobian matrix as a float array of shape (nE, nP, 1, 1), evaluated for
            each point in each cell. Default is None.
        dshp : Iterable, Optional
            The shape function derivatives of the master element as a float array of
            shape (nP, nNE, nDOF=6, 3), evaluated at a 'nP' number of points.
            Default is None.

        Returns
        -------
        numpy.ndarray
            The returned array has a shape of (nP, nNE=2, nDOF=6, 3), where nP, nNE and
            nDOF stand for the number of evaluation points, nodes per element and number of
            degrees of freedom, respectively. Number 3 refers to first, second and third
            derivatives.

        Notes
        -----
        The returned array is always 4 dimensional, even if there is only one
        evaluation point.
        """
        if pcoords is not None:
            lengths = self.lengths() if lengths is None else lengths
            # calculate derivatives wrt. the parametric coordinates in the range [-1, 1]
            pcoords = atleast1d(np.array(pcoords))
            rng = np.array([-1.0, 1.0]) if rng is None else np.array(rng)
            pcoords = to_range_1d(pcoords, source=rng, target=[-1.0, 1.0])
            dshp = self.__class__.dshpfnc(pcoords, lengths)
            if jac is None:
                # return derivatives wrt. the master frame
                return dshp.astype(float)
            else:
                # return derivatives wrt. the global frame
                return gdshpB(dshp, jac).astype(float)
        elif dshp is not None and jac is not None:
            # return derivatives wrt. the local frame
            return gdshpB(dshp, jac).astype(float)

    def shape_function_matrix(
        self,
        pcoords: Union[float, Iterable[float]] = None,
        *_,
        rng: Iterable = None,
        lengths: Iterable[float] = None,
        **__,
    ) -> ndarray:
        """
        Evaluates the shape function matrix at the points specified by 'pcoords'.

        Parameters
        ----------
        pcoords : float or Iterable
            Locations of the evaluation points.
        rng : Iterable, Optional
            The range in which the locations ought to be understood,
            typically [0, 1] or [-1, 1]. Default is [0, 1].

        Other Parameters
        ----------------
        These parameters are for advanced users and can be omitted.
        They are here only to avoid repeated evaulation of common quantities.

        lengths : Iterable, Optional
            The lengths of the beams in the block. Default is None.

        Returns
        -------
        numpy.ndarray
            The returned array has a shape of (nE, nP, nDOF, nDOF * nNE), where nE, nP,
            nNE and nDOF stand for the number of elements, evaluation points, nodes
            per element and number of degrees of freedom respectively.

        Notes
        -----
        The returned array is always 4 dimensional, even if there is only one
        evaluation point.
        """
        pcoords = atleast1d(np.array(pcoords))
        rng = np.array([-1.0, 1.0]) if rng is None else np.array(rng)
        pcoords = to_range_1d(pcoords, source=rng, target=[-1, 1])
        lengths = self.lengths() if lengths is None else lengths
        dshp = self.__class__.dshpfnc(pcoords, lengths)
        ecoords = self.local_coordinates()
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)
        shp = self.shape_function_values(pcoords, rng=rng)
        gdshp = self.shape_function_derivatives(jac=jac, dshp=dshp)
        return shape_function_matrix_Bernoulli_bulk(shp, gdshp).astype(float)

    def integrate_body_loads(self, values: ndarray) -> ndarray:
        """
        Returns nodal representation of body loads.

        Parameters
        ----------
        values : numpy.ndarray
            2d or 3d numpy float array of material densities of shape
            (nE, nNE * nDOF, nRHS) or (nE, nNE * nDOF), where nE, nNE, nDOF and nRHS
            stand for the number of elements, nodes per element, number of degrees
            of freedom and number of load cases respectively.

        Returns
        -------
        numpy.ndarray
            An array of shape (nE, nNE * 6, nRHS).

        Notes
        -----
        The returned array is always 3 dimensional, even if there is only one
        load case.

        See Also
        --------
        :func:`~body_load_vector_bulk`
        """
        values = atleastnd(values, 3, back=True).astype(float)
        # (nE, nNE * nDOF, nRHS) -> (nE, nRHS, nNE * nDOF)
        values = np.swapaxes(values, 1, 2)
        values = ascont(values)
        qpos, qweights = self.quadrature["full"]
        rng = np.array([-1.0, 1.0])
        shp = self.shape_function_values(qpos, rng=rng)
        # (nP, nNE=2, nDOF=6)
        dshp = self.shape_function_derivatives(qpos, rng=rng)
        # (nP, nNE=2, nDOF=6, 3)
        ecoords = self.local_coordinates()
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)
        # (nE, nP, 1, 1)
        djac = self.jacobian(jac=jac)  # (nE, nG)
        gdshp = self.shape_function_derivatives(jac=jac, dshp=dshp)
        # (nE, nP, nNE=2, nDOF=6, 3)
        return body_load_vector_Bernoulli(values, shp, gdshp, djac, qweights)

    def _internal_forces_H_(
        self,
        *_,
        cells: Union[int, Iterable[int]] = None,
        points: Union[float, Iterable] = None,  # [-1, 1]
    ) -> ndarray:
        shp = self.shape_function_values(points, rng=[-1, 1])[cells]
        dshp = self.shape_function_derivatives(points, rng=[-1, 1])[cells]
        ecoords = self.local_coordinates()[cells]
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)
        # jac -> (nE, nP, 1, 1)
        B = self.strain_displacement_matrix(shp=shp, dshp=dshp, jac=jac)
        # B -> (nE, nP, nSTRE, nNODE * 6)

        dofsol = self.dof_solution(flatten=True, cells=cells)
        # dofsol -> (nE, nNE * nDOF, nRHS)
        dofsol = ascont(np.swapaxes(dofsol, 1, 2))
        # dofsol -> (nE, nRHS, nEVAB)
        strains = approx_element_solution_bulk(dofsol, B)
        # strains -> (nE, nRHS, nP, nSTRE)
        strains -= self.kinetic_strains(points=points)[cells]
        D = self.elastic_material_stiffness_matrix()[cells]
        forces = calculate_internal_forces_bulk(strains, D)
        # forces -> (nE, nRHS, nP, nSTRE)
        forces = ascont(np.moveaxis(forces, 1, -1))
        dofsol = ascont(np.swapaxes(dofsol, 1, 2))
        # dofsol -> (nE, nEVAB, nRHS)
        # forces -> (nE, nP, nSTRE, nRHS)
        gdshp = self.shape_function_derivatives(jac=jac, dshp=dshp)
        nE, _, nRHS = dofsol.shape
        nNE, nDOF = self.__class__.NNODE, self.__class__.NDOFN
        dofsol = dofsol.reshape(nE, nNE, nDOF, nRHS)
        D = self.elastic_material_stiffness_matrix()[cells]
        forces = _postproc_bernoulli_internal_forces_H_(dofsol, forces, D, gdshp)
        return forces  # (nE, nP, 6, nRHS)

    def _internal_forces_L_(
        self,
        *_,
        cells: Union[int, Iterable[int]] = None,
        points: Union[float, Iterable] = None,  # [-1, 1]
    ) -> ndarray:
        local_points = np.array(self.lcoords()).flatten()

        shp = self.shape_function_values(local_points, rng=[-1, 1])[cells]
        dshp = self.shape_function_derivatives(local_points, rng=[-1, 1])[cells]
        ecoords = self.local_coordinates()[cells]
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)
        # jac -> (nE, nNE, 1, 1)
        B = self.strain_displacement_matrix(shp=shp, dshp=dshp, jac=jac)
        # B -> (nE, nNE, nSTRE, nNODE * 6)

        dofsol = self.dof_solution(flatten=True, cells=cells)
        # dofsol -> (nE, nNE * nDOF, nRHS)
        dofsol = ascont(np.swapaxes(dofsol, 1, 2))
        # dofsol -> (nE, nRHS, nEVAB)
        strains = approx_element_solution_bulk(dofsol, B)
        # strains -> (nE, nRHS, nNE, nSTRE)
        strains -= self.kinetic_strains(points=local_points)[cells]
        D = self.elastic_material_stiffness_matrix()[cells]

        nE, nRHS, nP = strains.shape[:3]
        forces = np.zeros((nE, nRHS, nP, 6), dtype=float)
        inds = np.array([0, 3, 4, 5], dtype=int)
        calculate_internal_forces_bulk(strains, D, forces, inds)
        # forces -> (nE, nRHS, nNE, nSTRE)
        forces = np.moveaxis(forces, 1, -1)
        # forces -> (nE, nNE, nSTRE, nRHS)

        *_, dshpf = self.Geometry.generate_class_functions(update=False, return_symbolic=False)
        dshp_geom = np.squeeze(dshpf([[i] for i in local_points]))
        # dshp_geom -> (nNE, nNE)
        _postproc_bernoulli_internal_forces_L_(forces, dshp_geom, jac)
        # forces -> (nE, nNE, nSTRE, nRHS)

        if isinstance(points, Iterable):
            approx = interpolate.interp1d(
                local_points, forces, axis=1, assume_sorted=True
            )
            forces = approx(points)

        return ascont(forces)

    def _internal_forces_(self, *args, **kwargs):
        return self._internal_forces_L_(self, *args, **kwargs)
        # return self._internal_forces_H_(self, *args, **kwargs)

    def lumped_mass_matrix(
        self,
        *_,
        lumping: str = "direct",
        alpha: float = 1 / 50,
        frmt: str = "full",
        **__,
    ) -> ndarray:
        """
        Returns the lumped mass matrix of the block.

        Parameters
        ----------
        alpha: float, Optional
            A nonnegative parameter, typically between 0 and 1/50 (see notes).
            Default is 1/20.
        lumping: str, Optional
            Controls lumping. Currently only direct lumping is available.
            Default is 'direct'.
        frmt: str, Optional
            Possible values are 'full' or 'diag'. If 'diag', only the diagonal
            entries are returned, if 'full' a full matrix is returned.
            Default is 'full'.
        """
        dbkey = self.__class__._attr_map_["M"]
        if lumping == "direct":
            dens = self.db.density
            try:
                areas = self.areas()
            except Exception:
                areas = np.ones_like(dens)
            lengths = self.lengths()
            topo = self.topology()
            ediags = dlump(dens, lengths, areas, topo, alpha)
            if frmt == "full":
                N = ediags.shape[-1]
                M = np.zeros((ediags.shape[0], N, N))
                inds = np.arange(N)
                M[:, inds, inds] = ediags
                self.db[dbkey] = M
                return M
            elif frmt == "diag":
                self.db[dbkey] = ediags
                return ediags
        raise RuntimeError("Lumping mode not recognized :(")



import sympy as sy


__all__ = ["smat_sym_ortho_3d", "cmat_sym_ortho_3d"]


imap_ref_ortho_3D = {0: (0, 0), 1: (1, 1), 2: (2, 2), 3: (1, 2), 4: (0, 2), 5: (0, 1)}


def upper(ij):
    return (ij[1], ij[0]) if ij[0] >= ij[1] else (ij[0], ij[1])


def smat_sym_ortho_3d(imap: dict = None, **subs) -> sy.Matrix:
    """
    Reurns the compliance matrix for a 3d orthotropic
    material in symbolic form.

    Parameters
    ----------
    imap : dict, Optional
        Index map. Default is None.
    subs : dict, Optional
        A dictionary of parameters to substitute.
    simplify : bool, Optional
        Whether to simplify the resulting symbolic matrix. Default is True.

    Notes
    -----
    The resulting symbolic matrix has 12 free symbols,
    but only 9 of these are independent.

    Returns
    -------
    sympy.Matrix
        6x6 symbolic compliance matrix.
    """
    E1, E2, E3 = sy.symbols("E1 E2 E3")
    G23, G12, G13 = sy.symbols("G23 G12 G13")
    NU12, NU21 = sy.symbols("NU12 NU21")
    NU13, NU31 = sy.symbols("NU13 NU31")
    NU23, NU32 = sy.symbols("NU23 NU32")
    S = sy.Matrix(
        [
            [1 / E1, -NU12 / E1, -NU13 / E1, 0, 0, 0],
            [-NU21 / E2, 1 / E2, -NU23 / E2, 0, 0, 0],
            [-NU31 / E3, -NU32 / E3, 1 / E3, 0, 0, 0],
            [0, 0, 0, 1 / G23, 0, 0],
            [0, 0, 0, 0, 1 / G13, 0],
            [0, 0, 0, 0, 0, 1 / G12],
        ]
    )
    if len(subs) > 0:
        S = S.subs([(sym, val) for sym, val in subs.items()])
    if imap is None:
        return S
    assert len(imap) == 6, "Indexmap must contain 6 indices."
    invmap = {upper(v): k for k, v in imap.items()}
    trmap = {i: invmap[imap_ref_ortho_3D[i]] for i in range(6)}
    S_new = sy.zeros(6, 6)
    for inds in trmap.items():
        S_new[inds[1], :] = _swap1d(S[inds[0], :], inds)
    return S_new


def cmat_sym_ortho_3d(imap: dict = None, **subs) -> sy.Matrix:
    """
    Reurns the stiffness matrix for a 3d orthotropic
    material in symbolic form.

    Parameters
    ----------
    imap : dict, optional
        Index map. Default is None.
    subs : dict, optional
        A dictionary of parameters to substitute.
    simplify : bool, Optional
        Whether to simplify the resulting symbolic matrix. Default is True.

    Notes
    -----
    The resulting symbolic matrix has 12 free symbols,
    but only 9 of these are independent.

    Returns
    -------
    sympy.Matrix
        6x6 symbolic stiffness matrix
    """
    res = smat_sym_ortho_3d(imap=imap, simplify=False, **subs).inv()
    res.simplify()
    return res


def _swap1d(a: sy.Matrix, inds: tuple):
    a.col_swap(*inds)
    return a


if __name__ == "__main__":
    C = cmat_sym_ortho_3d()



from polymesh.cells import TET10 as Tetra
from polymesh.utils.cells.gauss import Gauss_Legendre_Tet_1, Gauss_Legendre_Tet_4

from ..material.solid3d import Solid3d
from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


class TET10(ABC, Solid3d, Tetra, FiniteElement):
    qrule = "full"
    quadrature = {
        "full": Gauss_Legendre_Tet_4(),
        "selective": {(0, 1, 2): "full", (3, 4, 5): "reduced"},
        "reduced": Gauss_Legendre_Tet_1(),
    }



from polymesh.cells import Q9 as Quadrilateral
from polymesh.utils.cells.gauss import (
    Gauss_Legendre_Quad_9,
    Gauss_Legendre_Quad_4,
)

from ..material.membrane import Membrane
from ..material.mindlinplate import MindlinPlate
from ..material.mindlinshell import MindlinShell

from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


class Q9_M(ABC, Membrane, Quadrilateral, FiniteElement):
    qrule = "full"
    quadrature = {
        "full": Gauss_Legendre_Quad_9(),
        "selective": {(0, 1): "full", (2,): "reduced"},
        "reduced": Gauss_Legendre_Quad_4(),
    }


class Q9_P_MR(ABC, MindlinPlate, Quadrilateral, FiniteElement):
    qrule = "selective"
    quadrature = {
        "full": Gauss_Legendre_Quad_9(),
        "selective": {(0, 1, 2): "full", (3, 4): "reduced"},
        "reduced": Gauss_Legendre_Quad_4(),
    }


class Q9_S_MR(ABC, MindlinShell, Quadrilateral, FiniteElement):
    qrule = "selective"
    quadrature = {
        "full": Gauss_Legendre_Quad_9(),
        "selective": {(0, 1, 3, 4, 5): "full", (2, 6, 7): "reduced"},
        "reduced": Gauss_Legendre_Quad_4(),
    }



import numpy as np
from numpy import ndarray
from numba import njit, prange

__cache = True


@njit(nogil=True, cache=__cache)
def shape_function_values(x: float, L: float = 2) -> ndarray:
    """
    Evaluates the shape functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].
    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.
    """
    return np.array(
        [
            [
                0.5 - 0.5 * x,
                0.25 * x**3 - 0.75 * x + 0.5,
                0.25 * x**3 - 0.75 * x + 0.5,
                0.5 - 0.5 * x,
                -0.125 * L * (x - 1) ** 2 * (x + 1),
                0.125 * L * (x - 1) ** 2 * (x + 1),
            ],
            [
                0.5 * x + 0.5,
                -0.25 * x**3 + 0.75 * x + 0.5,
                -0.25 * x**3 + 0.75 * x + 0.5,
                0.5 * x + 0.5,
                -0.125 * L * (x - 1) * (x + 1) ** 2,
                0.125 * L * (x - 1) * (x + 1) ** 2,
            ],
        ]
    )


@njit(nogil=True, cache=__cache)
def shape_function_values_L(x: float) -> ndarray:
    """
    Evaluates the Lagrangian shape functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    Returns
    -------
    numpy.ndarray
        A 2d float array.
    """
    return np.array([0.5 - 0.5 * x, 0.5 * x + 0.5])


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_1(x: float, L: float = 2) -> ndarray:
    """
    Evaluates the first derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].
    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.
    """
    return np.array(
        [
            [
                -0.500000000000000,
                0.75 * x**2 - 0.75,
                0.75 * x**2 - 0.75,
                -0.500000000000000,
                L * (-0.375 * x**2 + 0.25 * x + 0.125),
                0.375 * L * (x - 1) * (1.0 * x + 0.333333333333333),
            ],
            [
                0.500000000000000,
                0.75 - 0.75 * x**2,
                0.75 - 0.75 * x**2,
                0.500000000000000,
                L * (-0.375 * x**2 - 0.25 * x + 0.125),
                0.375 * L * (1.0 * x - 0.333333333333333) * (x + 1),
            ],
        ]
    )


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_2(x: float, L: float = 2) -> ndarray:
    """
    Evaluates the second derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].
    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.
    """
    return np.array(
        [
            [0, 1.5 * x, 1.5 * x, 0, L * (0.25 - 0.75 * x), L * (0.75 * x - 0.25)],
            [0, -1.5 * x, -1.5 * x, 0, L * (-0.75 * x - 0.25), L * (0.75 * x + 0.25)],
        ]
    )


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_3(x: float, L: float = 2) -> ndarray:
    """
    Evaluates the third derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].
    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.
    """
    return np.array(
        [
            [0, 1.50000000000000, 1.50000000000000, 0, -0.75 * L, 0.75 * L],
            [0, -1.50000000000000, -1.50000000000000, 0, -0.75 * L, 0.75 * L],
        ]
    )


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_L(x: float) -> ndarray:
    """
    Evaluates the first derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    Returns
    -------
    numpy.ndarray
        A 2d float array.
    """
    return np.array([-0.500000000000000, 0.500000000000000])


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_values_bulk(x: ndarray, L: ndarray) -> ndarray:
    """
    Evaluates the shape functions at several points
    in the range [-1, 1].

    Parameters
    ----------
    x : numpy.ndarray
        1d array of floats in the range [-1, -1].
    L : numpy.ndarray
        Lengths of the beam elements. Default is None.

    Returns
    -------
    numpy.ndarray
        3d float array of shape (nP, nNE=2, nDOF=6).
    """
    nP = x.shape[0]
    nE = L.shape[0]
    res = np.zeros((nE, nP, 2, 6), dtype=x.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            res[iE, iP] = shape_function_values(x[iP], L[iE])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_values_multi_L(x: ndarray) -> ndarray:
    """
    Evaluates the Lagrangian shape functions at several points
    in the range [-1, 1].

    Parameters
    ----------
    x : numpy.ndarray
        1d array of floats in the range [-1, -1].

    Returns
    -------
    numpy.ndarray
        3d float array of shape (nP, nNE=2).
    """
    nP = x.shape[0]
    res = np.zeros((nP, 2), dtype=x.dtype)
    for iP in prange(nP):
        res[iP] = shape_function_values_L(x[iP])
    return res


@njit(nogil=True, cache=__cache)
def shape_function_derivatives(x: float, L: float = 2) -> ndarray:
    """
    Evaluates the derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].
    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        3d float array of shape (nNE=2, nDOF=6, 3).
    """
    res = np.zeros((2, 6, 3))
    res[:, :, 0] = shape_function_derivatives_1(x, L)
    res[:, :, 1] = shape_function_derivatives_2(x, L)
    res[:, :, 2] = shape_function_derivatives_3(x, L)
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_derivatives_bulk(x: ndarray, L: ndarray) -> ndarray:
    """
    Evaluates the derivatives of the shape
    functions at several points in the range [-1, 1].

    Parameters
    ----------
    x : numpy.ndarray
        1d array of floats in the range [-1, -1].
    L : numpy.ndarray
        Lengths of the beam elements. Default is None.

    Returns
    -------
    numpy.ndarray
        5d float array of shape (nE, nP, nNE=2, nDOF=6, 3).
    """
    nP = x.shape[0]
    nE = L.shape[0]
    res = np.zeros((nE, nP, 2, 6, 3), dtype=x.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            res[iE, iP] = shape_function_derivatives(x[iP], L[iE])
    return res


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_multi_L(x: ndarray) -> ndarray:
    """
    Evaluates the derivatives of the Lagrangian shape
    functions at multiple points x in the range [-1, 1].

    Parameters
    ----------
    x : numpy.ndarray
        1d array of floats in the range [-1, -1].

    Returns
    -------
    numpy.ndarray
        2d float array of shape (nP, nNE=2), where nP is the number
        of coordinates in 'x', nNE is the number of nodes of the cell.
    """
    nP = x.shape[0]
    res = np.zeros((nP, 2), dtype=x.dtype)
    for iP in prange(nP):
        res[iP] = shape_function_derivatives_L(x[iP])
    return res



"""
The ANDES template.
"""
import numpy as np
from numpy import ndarray
from numba import njit, prange

from neumann.utils import repeat1d
from polymesh.cells import T3 as Triangle
from polymesh.utils.tri import lcoords_tri, shp_tri_loc

from ..material.membrane import DrillingMembrane

from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC

__cache = True


__all__ = ["LSTMR"]


def sig_ANDES(key: str = "OPT", C: ndarray = None):
    if key == "OPT":
        # OPTIMAL ANDES TEMPLATE
        sig = sig_ANDES_opt(C)
    elif key == "ALL88-3I":
        # Allman 88 element integrated by 3-point interior rule
        sig = (
            1,
            4 / 9,
            1 / 12,
            5 / 12,
            1 / 2,
            0,
            1 / 3,
            -1 / 3,
            -1 / 12,
            -1 / 2,
            -5 / 12,
        )
    elif key == "ALL88-3M":
        # Allman 88 element integrated by 3 midpoint rule
        sig = (1, 4 / 9, 1 / 4, 5 / 4, 3 / 2, 0, 1, -1, -1 / 4, -3 / 2, -5 / 4)
    elif key == "ALL88-LS":
        # Allman 88 element least-square strain fit
        sig = (
            1,
            4 / 9,
            3 / 20,
            3 / 4,
            9 / 10,
            0,
            3 / 5,
            -3 / 5,
            -3 / 20,
            -9 / 10,
            -3 / 4,
        )
    elif key == "LST-Ret":
        # Retrofitted LST with alpha=4/3
        sig = (4 / 3, 1 / 2, 2 / 3, -2 / 3, 0, 0, -4 / 3, 4 / 3, -2 / 3, 0, 2 / 3)
    elif key == "CST":
        # Constant strain triangle CST-3/6C
        sig = (0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    else:
        raise ValueError("Invalid signature.")
    return np.array(sig).astype(float)


@njit(nogil=True, cache=__cache)
def sig_ANDES_opt(C: ndarray):
    E11, E12, E13, _, E22, E23, _, _, E33 = C.flatten()
    Edet = (
        E11 * E22 * E33
        + 2 * E12 * E13 * E23
        - E11 * E23**2
        - E22 * E13**2
        - E33 * E12**2
    )
    W = (
        -6 * E12**3
        + 5 * E11**2 * E22
        - 5 * E12**2 * E22
        - E22 * (75 * E13**2 + 14 * E13 * E23 + 3 * E23**2)
        + 2 * E12 * (7 * E13**2 + 46 * E13 * E23 + 7 * E23**2)
        - E11
        * (
            5 * E12**2
            + 3 * E13**2
            - 6 * E12 * E22
            - 5 * E22**2
            + 14 * E13 * E23
            + 75 * E23**2
        )
        + (
            3 * E11**2
            + 82 * E11 * E22
            + 3 * E22**2
            - 4 * (6 * E12**2 + 5 * E13**2 - 6 * E13 * E23 + 5 * E23**2)
        )
        * E33
        + 4 * (5 * E11 - 6 * E12 + 5 * E22) * E33**2
    )
    E11C11avg = W / (128 * Edet)
    b0 = max([2 / E11C11avg - 3 / 2, 1 / 100])
    return np.array([3 / 2, b0, 1, 2, 1, 0, 1, -1, -1, -1, -2], dtype=C.dtype)


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def sig_ANDES_opt_bulk(C: ndarray):
    nE = len(C)
    res = np.zeros((nE, 11), dtype=C.dtype)
    for i in prange(nE):
        res[i] = sig_ANDES_opt(C[i])
    return res


@njit(nogil=True, cache=__cache)
def _stiffness_data_(ecoords, h, sig):
    (x1, x2, x3), (y1, y2, y3) = ecoords[:, 0], ecoords[:, 1]
    ab, b0, b1, b2, b3, b4, b5, b6, b7, b8, b9 = sig
    x12, x21, x13, x31, x23, x32 = x1 - x2, x2 - x1, x1 - x3, x3 - x1, x2 - x3, x3 - x2
    y12, y21, y13, y31, y23, y32 = y1 - y2, y2 - y1, y1 - y3, y3 - y1, y2 - y3, y3 - y2
    LL21 = x21**2 + y21**2
    LL32 = x32**2 + y32**2
    LL13 = x13**2 + y13**2
    A = (y21 * x13 - x21 * y13) / 2
    A2, A4 = 2 * A, 4 * A
    L = (
        np.array(
            [
                [y23, 0, x32],
                [0, x32, y23],
                [
                    ab * y23 * (y13 - y21) / 6,
                    ab * x32 * (x31 - x12) / 6,
                    ab * (x31 * y13 - x12 * y21) / 3,
                ],
                [y31, 0, x13],
                [0, x13, y31],
                [
                    ab * y31 * (y21 - y32) / 6,
                    ab * x13 * (x12 - x23) / 6,
                    ab * (x12 * y21 - x23 * y32) / 3,
                ],
                [y12, 0, x21],
                [0, x21, y12],
                [
                    ab * y12 * (y32 - y13) / 6,
                    ab * x21 * (x23 - x31) / 6,
                    ab * (x23 * y32 - x31 * y13) / 3,
                ],
            ],
            dtype=ecoords.dtype,
        )
        * h
        / 2
    )
    Tru = (
        np.array(
            [
                [x32, y32, A4, x13, y13, 0, x21, y21, 0],
                [x32, y32, 0, x13, y13, A4, x21, y21, 0],
                [x32, y32, 0, x13, y13, 0, x21, y21, A4],
            ],
            dtype=ecoords.dtype,
        )
        / A4
    )
    Te = np.array(
        [
            [y23 * y13 * LL21, y31 * y21 * LL32, y12 * y32 * LL13],
            [x23 * x13 * LL21, x31 * x21 * LL32, x12 * x32 * LL13],
            [
                (y23 * x31 + x32 * y13) * LL21,
                (y31 * x12 + x13 * y21) * LL32,
                (y12 * x23 + x21 * y32) * LL13,
            ],
        ],
        dtype=ecoords.dtype,
    ) / (4 * A**2)
    Q1 = (
        np.array(
            [
                [b1 / LL21, b2 / LL21, b3 / LL21],
                [b4 / LL32, b5 / LL32, b6 / LL32],
                [b7 / LL13, b8 / LL13, b9 / LL13],
            ],
            dtype=ecoords.dtype,
        )
        * A2
        / 3
    )
    Q2 = (
        np.array(
            [
                [b9 / LL21, b7 / LL21, b8 / LL21],
                [b3 / LL32, b1 / LL32, b2 / LL32],
                [b6 / LL13, b4 / LL13, b5 / LL13],
            ],
            dtype=ecoords.dtype,
        )
        * A2
        / 3
    )
    Q3 = (
        np.array(
            [
                [b5 / LL21, b6 / LL21, b4 / LL21],
                [b8 / LL32, b9 / LL32, b7 / LL32],
                [b2 / LL13, b3 / LL13, b1 / LL13],
            ],
            dtype=ecoords.dtype,
        )
        * A2
        / 3
    )
    return A, L, Te, b0, Q1, Q2, Q3, Tru


@njit(nogil=True, cache=__cache)
def _stiffness_matrix_LSTMR(C, ecoords, h, sig):
    A, L, Te, b0, Q1, Q2, Q3, Tru = _stiffness_data_(ecoords, h, sig)
    V = h * A
    Q4, Q5, Q6 = (Q1 + Q2) / 2, (Q3 + Q2) / 2, (Q1 + Q3) / 2
    Kb = (L @ C @ L.T) / V
    Cn = Te.T @ (C @ Te)
    Kr = (3 * b0 * V / 4) * (Q4.T @ Cn @ Q4 + Q5.T @ Cn @ Q5 + Q6.T @ Cn @ Q6)
    Kh = Tru.T @ Kr @ Tru
    return Kb, Kh


@njit(nogil=True, cache=__cache)
def __stiffness_matrix__(C, ecoords, h, sig):
    Kb, Kh = _stiffness_matrix_LSTMR(C, ecoords, h, sig)
    return Kb + Kh


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix(C, ecoords, h, sig):
    nE = len(C)
    res = np.zeros((nE, 9, 9), dtype=C.dtype)
    for i in prange(nE):
        res[i, :, :] = __stiffness_matrix__(C[i], ecoords[i], h[i], sig[i])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix_LSQ_4_12(
    C: ndarray,
    ecoords: ndarray,
    h: ndarray,
    sig: ndarray,
    dmaps: ndarray,
    subtopo: ndarray,
):
    nE = len(C)
    dmax = dmaps.max() + 1
    res = np.zeros((nE, dmax, dmax), dtype=C.dtype)
    for i in range(subtopo.shape[0]):
        K_i_bulk = stiffness_matrix(C, ecoords[:, subtopo[i], :], h, sig)
        for j in prange(K_i_bulk.shape[1]):
            for k in prange(K_i_bulk.shape[2]):
                res[:, dmaps[i, j], dmaps[i, k]] += K_i_bulk[:, j, k]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def approx_strains(pcoord, ecoords, h, sig, dofsol1d, gnum):
    nE = len(gnum)
    res = np.zeros((nE, 3), dtype=ecoords.dtype)
    A1, A2, A3 = shp_tri_loc(pcoord)
    for i in prange(nE):
        A, L, Te, _, Q1, Q2, Q3, Tru = _stiffness_data_(ecoords[i], h[i], sig[i])
        uE = dofsol1d[gnum]
        res[i, :] = (1 / (h[i] * A)) * (L.T @ uE) + 3 * (
            Te @ (A1 * Q1 + A2 * Q2 + A3 * Q3) @ Tru @ uE
        ) / 2
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def strain_displacement_matrix(pcoord, ecoords, h, sig):
    nE = ecoords.shape[0]
    res = np.zeros((nE, 3, 9), dtype=ecoords.dtype)
    A1, A2, A3 = shp_tri_loc(pcoord)
    for i in prange(nE):
        A, L, Te, _, Q1, Q2, Q3, Tru = _stiffness_data_(ecoords[i], h[i], sig[i])
        res[i, :] = (1 / (h[i] * A)) * L.T + 3 * (
            Te @ (A1 * Q1 + A2 * Q2 + A3 * Q3) @ Tru
        ) / 2
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def approx_stresses(pcoord, C, ecoords, h, sig, dofsol1d, gnum):
    e = approx_strains(pcoord, ecoords, h, sig, dofsol1d, gnum)
    nE = len(e)
    res = np.zeros((nE, 3), dtype=C.dtype)
    for i in prange(nE):
        res[i, :] = C[i] @ e[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stresses_at_nodes(C, ecoords, h, sig, dofsol1d, gnum):
    lcoords = lcoords_tri()
    nE = len(C)
    res = np.zeros((nE, 3, 3), dtype=C.dtype)
    for i in prange(3):
        res[:, i, :] = approx_stresses(lcoords[i], C, ecoords, h, sig, dofsol1d, gnum)
    return res


class T3_ALL88_3I(ABC, DrillingMembrane, Triangle, FiniteElement):
    """
    Allman 88 element integrated by 3-point interior rule, constructed
    using the ANDES template of Carmelo Militello and Carlos A. Felippa.

    References
    ----------
    .. [1] Carlos A. Felippa, A study of optimal membrane triangles with
       drilling freedoms, Computer Methods in Applied Mechanics and Engineering,
       Volume 192, Issues 16-18, 2003, Pages 2125-2168, ISSN 0045-7825,
       https://doi.org/10.1016/S0045-7825(03)00253-6.
    """

    version = "ALL88-3I"
    qrule = "full"
    quadrature = {
        "full": (np.array([[1 / 3, 1 / 3]]), np.array([1 / 2])),
    }

    def _elastic_stiffness_matrix_(
        self, *, transform: bool = True, **kwargs
    ) -> ndarray:
        ec = kwargs.get("_ec", None)
        if ec is None:
            _frames = kwargs.get("_frames", self.frames)
            if _frames is not None:
                ec = self.local_coordinates(target=_frames)
            else:
                _topo = kwargs.get("_topo", self.topology().to_numpy())
                ec = self.points_of_cells(topo=_topo)
        D = self.model_stiffness_matrix()
        sig = np.array(sig_ANDES(key=self.version), dtype=float)
        sig = repeat1d(sig, D.shape[0]).reshape(D.shape[0], 11)
        h = np.ones((D.shape[0]), dtype=float)
        K = stiffness_matrix(D, ec, h, sig)
        return self._transform_coeff_matrix_(K) if transform else K

    @classmethod
    def strain_displacement_matrix(cls, pcoord, ecoords):
        # FIXME : not tested, probably wrong
        nE = ecoords.shape[0]
        sig = np.array(sig_ANDES(key=cls.version), dtype=float)
        sig = repeat1d(sig, nE).reshape(nE, 11)
        h = np.ones((nE), dtype=float)
        return strain_displacement_matrix(pcoord, ecoords, h, sig)

    def _strain_load_vector_(self, values: ndarray) -> ndarray:
        # FIXME : returns a zero, but shape-correct array
        nE, _, nRHS = values.shape
        nNE = self.NNODE
        nDOF = self.NDOFN
        nodal_loads = np.zeros((nE, nNE * nDOF, nRHS))
        return nodal_loads  # (nE, nTOTV, nRHS)



from polymesh.config import __hasplotly__, __hasmatplotlib__

from .mesh import FemMesh

from .cells import Q4_M, Q9_M


class SurfaceMesh(FemMesh):
    """
    A data class dedicated to 2d cells. It handles sections and other line
    related information, presets for plotting, etc.

    """

    _cell_classes_ = {
        4: Q4_M,
        9: Q9_M,
    }

    def __init__(self, *args, section=None, **kwargs):
        if section is not None:
            pass
        self._section = section
        super().__init__(*args, **kwargs)

    @property
    def section(self):
        if self._section is not None:
            return self._section
        else:
            if self.is_root():
                return self._section
            else:
                return self.parent.section



from .mesh import FemMesh, SolidMesh, MembraneMesh
from .linemesh import LineMesh, BernoulliFrame
from .pointdata import PointData
from .cells.celldata import CellData
from .structure import Structure
from .ebc import NodalSupport, NodeToNode



# -*- coding: utf-8 -*-
import codecs
import os.path
from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def get_description(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__description__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find description string.")


with open("README.md", "r") as fh:
    long_description = fh.read()


with open("requirements.txt") as f:
    required = f.read().splitlines()


_init_path = "src/sigmaepsilon/__init__.py"
_version = get_version(_init_path)
_description = get_description(_init_path)
_url = "https://github.com/dewloosh/sigmaepsilon"
_download_url = _url + "/archive/refs/tags/{}.zip".format(_version)


setup(
    name="sigmaepsilon",
    version=_version,
    author="Bence Balogh",
    author_email="benceeok@gmail.com",
    description=_description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=_url,
    download_url=_download_url,
    packages=find_packages(where="src"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: Free for non-commercial use",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    python_requires=">=3.6, <3.11",
    package_dir={"": "src"},
    install_requires=required,
    zip_safe=False,
)



from polymesh.cells import H8 as HexaHedron
from polymesh.utils.cells.gauss import Gauss_Legendre_Hex_Grid

from ..material.solid3d import Solid3d
from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


class H8(ABC, Solid3d, HexaHedron, FiniteElement):
    qrule = "selective"
    quadrature = {
        "full": Gauss_Legendre_Hex_Grid(2, 2, 2),
        "selective": {(0, 1, 2): "full", (3, 4, 5): "reduced"},
        "reduced": Gauss_Legendre_Hex_Grid(1, 1, 1),
    }



from polymesh.cells import T3 as Triangle

from ..material.membrane import Membrane
from ..material.mindlinplate import MindlinPlate
from ..material.mindlinshell import MindlinShell

from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


class CST_M(ABC, Membrane, Triangle, FiniteElement):
    """
    The constant-strain triangle a.k.a., CST triangle, Turner triangle or
    linear triangle for membranes. Developed as a plane stress element by
    John Turner, Ray Clough and Harold Martin in 1952-53 [1], published in 1956 [2].

    Notes
    -----
    The element has poor performance and is represented for historycal reasons.
    Don't use it in a production enviroment, unless your mesh is extremely dense.

    References
    ----------
    .. [1] R. W. Clough, The finite element method - a personal view of its original
       formulation, in From Finite Elements to the Troll Platform - the Ivar Holand
       70th Anniversary Volume, ed. by K. Bell, Tapir, Trondheim, Norway, 89-100, 1994.
    .. [2] M. J. Turner, R. W. Clough, H. C. Martin, and L. J. Topp, Stiffness and
       deflection analysis of complex structures, J. Aero. Sco., 23, pp. 805-824, 1956.
    """

    qrule = "full"


class CST_P_MR(ABC, MindlinPlate, Triangle, FiniteElement):
    """
    The constant-strain triangle (a.k.a., CST triangle, Turner triangle)
    """

    qrule = "full"


class CST_S_MR(ABC, MindlinShell, Triangle, FiniteElement):
    qrule = "full"



from typing import Iterable

import numpy as np


def points_to_rectangle_region(points: np.ndarray):
    xmin = points[:, 0].min()
    ymin = points[:, 1].min()
    xmax = points[:, 0].max()
    ymax = points[:, 1].max()
    return xmin, ymin, xmax - xmin, ymax - ymin


def rectangle_coordinates(
    *_,
    region: Iterable = None,
    xy: Iterable = None,
    w: float = None,
    h: float = None,
    center: Iterable = None
) -> Iterable:
    """
    Returns the bottom-left and upper-right coordinates of the region
    of the load from several inputs.

    Parameters
    ----------
    region : Iterable, Optional
        An iterable of length 4 with values x0, y0, w, and h. Here x0 and y0 are
        the coordinates of the bottom-left corner, w and h are the width and height
        of the region.
    xy : Iterable, Optional
        The position of the bottom-left corner as an iterable of length 2.
    w : float, Optional
        The width of the region.
    h : float, Optional
        The height of the region.
    center : Iterable, Optional
        The coordinates of the center of the region.

    Returns
    -------
    numpy.ndarray
        A 2d float array of coordinates, where the entries of the first and second
        rows are the coordinates of the lower-left and upper-right points of the region.

    Examples
    --------
    The following definitions return the same output:

    >>> from sigmaepsilon.fourier.utils import rectangle_coordinates
    >>> rectangle_coordinates(region=[2, 3, 0.5, 0.7])
    >>> rectangle_coordinates(xy=[2, 3], w=0.5, h=0.7)
    >>> rectangle_coordinates(center=[2.25, 3.35], w=0.5, h=0.7)
    """
    points = None
    if region:
        x0, y0, w, h = np.array(region)
        points = np.array([[x0, y0], [x0 + w, y0 + h]])
    elif xy and w and h:
        (x0, y0), w, h = xy, w, h
        points = np.array([[x0, y0], [x0 + w, y0 + h]])
    elif center and w and h:
        (xc, yc), w, h = center, w, h
        points = np.array([[xc - w / 2, yc - h / 2], [xc + w / 2, yc + h / 2]])
    else:
        ValueError("Invalid definition!")
    return points


def sin1d(x, i=1, L=1):
    return np.sin(x * np.pi * i / L)


def cos1d(x, i=1, L=1):
    return np.cos(x * np.pi * i / L)



from enum import Enum
from typing import Iterable, List


class DOF(Enum):
    """
    An enumeration class to handle degrees of freedom.
    It acceprs several keys for the same DOF to help dissolve
    ambiguities between different notations.

    Examples
    --------
    >>> from sigmaepsilon.fem.config import DOF
    >>> DOF['UXX']
    <DOF.ROTX: 3>
    >>> DOF.UXX
    <DOF.ROTX: 3>
    >>> DOF.UXX.value
    3
    >>> DOF.UXX.name
    'ROTX'

    The keys 'UX', 'U1' and 'U' all refer to the same thing:

    >>> DOF.dofmap(['U', 'UX', 'U1'])
    [0, 0, 0]

    Similarly, the rotation around X can be referenced using
    multiple keys (for rotations, a total of 7 versions are available):

    >>> DOF.dofmap(['ROTX', 'UXX', 'UYZ', 'U32', 'U23'])
    [3, 3, 3, 3, 3]

    """

    UX = U1 = U = 0  # displacement in X direction
    UY = U2 = V = 1  # displacement in Y direction
    UZ = U3 = W = 2  # displacement in Z direction
    ROTX = UXX = UYZ = UZY = U11 = U23 = U32 = 3  # rotation around X
    ROTY = UYY = UXZ = UZX = U22 = U13 = U31 = 4  # rotation around Y
    ROTZ = UZZ = UXY = UYX = U33 = U12 = U21 = 5  # rotation around Z

    @classmethod
    def dofmap(cls, dofs: Iterable) -> List:
        """
        Returns indices of dofs.

        Examples
        --------
        >>> from sigmaepsilon.fem.config import DOF
        >>> DOF.dofmap(['UX', 'U', 'UXX'])
        [0, 0, 3]
        """
        return [cls[d].value for d in dofs]


if __name__ == "__main__":
    DOF.UXX, DOF["UXX"]
    type(DOF["UXX"].value)



from polymesh.utils.cells.q4 import area_Q4_bulk
from polymesh.cells import Q4 as Quadrilateral
from polymesh.utils.cells.gauss import (
    Gauss_Legendre_Quad_1,
    Gauss_Legendre_Quad_4,
)

from ..material.membrane import Membrane
from ..material.mindlinplate import MindlinPlate
from ..material.mindlinshell import MindlinShell

from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


class Q4_M(ABC, Membrane, Quadrilateral, FiniteElement):
    qrule = "full"
    quadrature = {
        "full": Gauss_Legendre_Quad_4(),
        "selective": {(0, 1): "full", (2,): "reduced"},
        "reduced": Gauss_Legendre_Quad_1(),
    }

    def areas(self, *args, **kwargs):
        """This shadows the original geometrical implementation."""
        topo = self.topology().to_numpy() if topo is None else topo
        return area_Q4_bulk(self.local_coordinates(topo=topo))


class Q4_P_MR(ABC, MindlinPlate, Quadrilateral, FiniteElement):
    qrule = "selective"
    quadrature = {
        "full": Gauss_Legendre_Quad_4(),
        "selective": {(0, 1, 2): "full", (3, 4): "reduced"},
        "reduced": Gauss_Legendre_Quad_1(),
    }


class Q4_S_MR(ABC, MindlinShell, Quadrilateral, FiniteElement):
    qrule = "selective"
    quadrature = {
        "full": Gauss_Legendre_Quad_4(),
        "selective": {(0, 1, 3, 4, 5): "full", (2, 6, 7): "reduced"},
        "reduced": Gauss_Legendre_Quad_1(),
    }



import numpy as np
from typing import Tuple, Union, Iterable

from linkeddeepdict import LinkedDeepDict
from neumann import atleast2d

from .problem import NavierProblem
from .loads import LoadGroup, NavierLoadError
from .preproc import lhs_Navier
from .postproc import postproc
from .proc import linsolve_Kirchhoff, linsolve_Mindlin


class RectangularPlate(NavierProblem):
    """
    A class to handle semi-analytic solutions of rectangular plates with
    specific boudary conditions.

    Parameters
    ----------
    size : Tuple[float]
        The size of the rectangle.
    shape : Tuple[int]
        Numbers of harmonic terms involved in both directions.
    """

    def __init__(
        self,
        size: Tuple[float],
        shape: Tuple[int],
        *,
        D11: float = None,
        D12: float = None,
        D22: float = None,
        D66: float = None,
        S44: float = None,
        S55: float = None,
        D: Iterable = None,
        S: Iterable = None,
    ):
        super().__init__()
        self.size = np.array(size, dtype=float)
        self.shape = np.array(shape, dtype=int)

        if D11 and D12 and D22 and D66:
            D = np.zeros((3, 3))
            D[0, 0] = self.D11
            D[0, 1] = self.D12
            D[1, 0] = self.D12
            D[1, 1] = self.D22
            D[2, 2] = self.D66
            self.D = D
        else:
            self.D = np.array(D, dtype=float)

        if S44 and S55:
            S = np.zeros((2, 2))
            S[0, 0] = self.S44
            S[1, 1] = self.S55
            self.S = S
        elif S is not None:
            self.S = np.array(S, dtype=float)
        else:
            self.S = None

    def solve(self, loads: Union[dict, LoadGroup], points: Iterable) -> LinkedDeepDict:
        """
        Solves the problem and calculates all entities at the specified points.

        Parameters
        ----------
        loads : Union[dict, LoadGroup]
            The loads.
        points : Iterable
            2d float array of coordinates, where the results are to be evaluated.

        Returns
        -------
        dict
            A dictionary with a same layout as the input.
        """
        # STIFFNESS
        LHS = lhs_Navier(self.size, self.shape, D=self.D, S=self.S)

        # LOADS
        if isinstance(loads, LoadGroup):
            _loads = loads
        elif isinstance(loads, dict):
            _loads = LoadGroup.from_dict(loads)
        else:
            raise NavierLoadError()
        _loads.problem = self
        LC = list(_loads.cases())
        RHS = np.vstack(list(lc.rhs() for lc in LC))

        # SOLUTION
        if self.S is None:
            """_RHS = rhs_Kirchhoff(RHS, self.length)
            coeffs = linsolve_Kirchhoff(LHS, _RHS)
            del _RHS"""
            # (nLHS, nRHS, nMN)
        else:
            coeffs = linsolve_Mindlin(LHS, RHS)
            # (nLHS, nRHS, nMN, 3)

        # POSTPROCESSING
        points = atleast2d(points)
        res = postproc(self.size, self.shape, points, coeffs, RHS, self.D, self.S)
        # (nLHS, nRHS, nP, nX)
        result = LinkedDeepDict()
        for i, lc in enumerate(LC):
            result[lc.address] = res[0, i, :, :]
        result.lock()
        return result



import unittest
from neumann.linalg import Vector, linspace
from neumann import repeat, minmax
from neumann.logical import allclose
from polymesh.space import StandardFrame, PointCloud
from polymesh.utils.space import index_of_closest_point, index_of_furthest_point
from polymesh.utils.topology import L2_to_L3
from sigmaepsilon.fem.cells import B2, B3
from sigmaepsilon import Structure, LineMesh, PointData
from sigmaepsilon.fem import NodalSupport
from sigmaepsilon.utils.fem.dyn import Rayleigh_quotient, effective_modal_masses
import numpy as np
from numpy import pi as PI


class TestBernoulliLinearStatics(unittest.TestCase):
    def check_relative_error(self, v_analytic, v_fem, tol: float = 0.1):
        """
        Compares the absolute relative error between the analytic and fem
        results against the given tolerance. The tolerance is understood
        as maximum percentage.
        """
        if v_analytic < 0:
            v_analytic -= 1
            v_fem -= 1
        else:
            v_analytic += 1
            v_fem += 1
        return np.abs(100 * (v_analytic - v_fem) / v_analytic) < tol

    def test_conc_tr(self):
        """
        Straight console, with starting point at (0, 0, 0), length L, and
        random orientation of axes. The same concentrated forces and moments
        are applied at the free end in the coordinate system of the console.
        The results should be the same when understood in the frame of the
        console, independently of the random orientation angles.
        """
        L, d, t = 1000.0, 100.0, 5.0  # geometry
        Ex, nu = 210000.0, 0.25  # material
        penalty = 1e20  # penalty value for essential BCs
        np.random.RandomState()
        # section
        w, h = 5.0, 100.0  # width and height of the rectangular cross section
        A = w * h  # area
        Iy = w * h**3 / 12  # second moment of inertia around the y axis
        Iz = h * w**3 / 12  # second moment of inertia around the z axis
        Ix = Iy + Iz  # torsional inertia
        # material
        G = Ex / (2 * (1 + nu))
        Hooke = np.array(
            [
                [Ex * A, 0, 0, 0],
                [0, G * Ix, 0, 0],
                [0, 0, Ex * Iy, 0],
                [0, 0, 0, Ex * Iz],
            ]
        )

        def solve(n, angles, loads, celltype=B2):
            # space
            GlobalFrame = StandardFrame(dim=3)
            TargetFrame = GlobalFrame.rotate("Body", angles, "XYZ", inplace=False)
            # mesh
            p0 = np.array([0.0, 0.0, 0.0])
            p1 = np.array([L, 0.0, 0.0])
            coords = linspace(p0, p1, n + 1)
            points = PointCloud(coords, frame=TargetFrame)
            coords = points.show(GlobalFrame)
            topo = np.zeros((n, 2), dtype=int)
            topo[:, 0] = np.arange(n)
            topo[:, 1] = np.arange(n) + 1
            if celltype.NNODE == 3:
                coords, topo = L2_to_L3(coords, topo)
            i_first = index_of_closest_point(coords, np.array([0.0, 0.0, 0.0]))
            i_last = index_of_furthest_point(coords, np.array([0.0, 0.0, 0.0]))
            # essential boundary conditions
            fixity = np.zeros((coords.shape[0], 6)).astype(bool)
            fixity[i_first, :] = True
            fixity = fixity.astype(float) * penalty
            # natural boundary conditions
            loads = np.array(loads)
            nodal_loads = np.zeros((coords.shape[0], 6))
            vF = Vector(loads[:3], frame=TargetFrame).show(GlobalFrame)
            vM = Vector(loads[3:], frame=TargetFrame).show(GlobalFrame)
            nodal_loads[i_last, :3] = vF
            nodal_loads[i_last, 3:] = vM
            # pointdata
            pd = PointData(
                coords=coords, frame=GlobalFrame, loads=nodal_loads, fixity=fixity
            )
            # celldata
            frames = repeat(TargetFrame.axes, topo.shape[0])
            cd = celltype(topo=topo, material=Hooke, frames=frames)
            # set up mesh and structure
            mesh = LineMesh(pd, cd, frame=GlobalFrame)
            structure = Structure(mesh=mesh)
            structure.linear_static_analysis()
            dofsol = structure.nodal_dof_solution(store="dofsol")
            u = np.zeros(6)
            u[:3] = Vector(dofsol[i_last, :3], frame=GlobalFrame).show(TargetFrame)
            u[3:] = Vector(dofsol[i_last, 3:], frame=GlobalFrame).show(TargetFrame)
            reactions = structure.reaction_forces()
            r = np.zeros(6)
            r[:3] = Vector(reactions[i_first, :3], frame=GlobalFrame).show(TargetFrame)
            r[3:] = Vector(reactions[i_first, 3:], frame=GlobalFrame).show(TargetFrame)
            forces = structure.internal_forces()
            f = np.zeros(6)
            f[:3] = forces[0, 0, :3, 0]
            f[3:] = forces[0, 0, 3:, 0]
            return u, r, f

        # ######################################################
        # CONCENTRATED FORCES AT THE FREE END
        # ######################################################

        Fx, Fy, Fz = 1.0, 1.0, 1.0
        loads = [Fx, Fy, Fz, 0, 0, 0]
        n = 20  # number of elements
        cells = [B2, B3]
        # analytical solutions
        UX = Fx * L / (Ex * A)  # displacement at the free end
        UY = Fy * L**3 / (3 * Ex * Iz)  # displacement at the free end
        UZ = Fz * L**3 / (3 * Ex * Iy)  # displacement at the free end
        UXX = 0.0  # rotation at the free end
        UYY = -Fz * L**2 / (2 * Ex * Iy)  # rotation at the free end
        UZZ = Fy * L**2 / (2 * Ex * Iz)  # rotation at the free end
        RX = -Fx  # reaction force
        RY = -Fy  # reaction force
        RZ = -Fz  # reaction force
        RXX = 0.0  # reaction moment
        RYY = Fz * L  # reaction moment
        RZZ = -Fy * L  # reaction moment
        FX = Fx  # internal force at the support
        FY = Fy  # internal force at the support
        FZ = Fz  # internal force at the support
        FXX = 0.0  # internal moment at the support
        FYY = -Fz * L  # internal moment at the support
        FZZ = Fy * L  # internal moment at the support
        for cell in cells:
            for _ in range(3):
                angles = np.random.rand(3) * np.pi * 2
                u, r, f = solve(n, angles, loads, cell)
                UX_fem, UXX_fem = u[0], u[3]
                RX_fem, RXX_fem = r[0], r[3]
                FX_fem, FXX_fem = f[0], f[3]
                UY_fem, UZZ_fem = u[1], u[5]
                RY_fem, RZZ_fem = r[1], r[5]
                FY_fem, FZZ_fem = f[1], f[5]
                UZ_fem, UYY_fem = u[2], u[4]
                RZ_fem, RYY_fem = r[2], r[4]
                FZ_fem, FYY_fem = f[2], f[4]
                assert self.check_relative_error(UX, UX_fem)
                assert self.check_relative_error(UY, UY_fem)
                assert self.check_relative_error(UZ, UZ_fem)
                assert self.check_relative_error(UXX, UXX_fem)
                assert self.check_relative_error(UYY, UYY_fem)
                assert self.check_relative_error(UZZ, UZZ_fem)
                assert self.check_relative_error(RX, RX_fem)
                assert self.check_relative_error(RY, RY_fem)
                assert self.check_relative_error(RZ, RZ_fem)
                assert self.check_relative_error(RXX, RXX_fem)
                assert self.check_relative_error(RYY, RYY_fem)
                assert self.check_relative_error(RZZ, RZZ_fem)
                assert self.check_relative_error(FX, FX_fem)
                assert self.check_relative_error(FY, FY_fem)
                assert self.check_relative_error(FZ, FZ_fem)
                assert self.check_relative_error(FXX, FXX_fem)
                assert self.check_relative_error(FYY, FYY_fem)
                assert self.check_relative_error(FZZ, FZZ_fem)

        # ######################################################
        # CONCENTRATED MOMENTS AT THE FREE END
        # ######################################################

        Mx, My, Mz = 1.0, 1.0, 1.0
        loads = [0, 0, 0, Mx, My, Mz]
        n = 2  # number of elements
        cells = [B2, B3]
        # analytical solutions
        UX = 0.0  # displacement at the free end
        UY = Mz * L**2 / (2 * Ex * Iz)  # displacement at the free end
        UZ = -My * L**2 / (2 * Ex * Iy)  # displacement at the free end
        UXX = Mx * L / (G * Ix)  # rotation at the free end
        UYY = My * L / (Ex * Iy)  # rotation at the free end
        UZZ = Mz * L / (Ex * Iz)  # rotation at the free end
        RX = 0.0  # reaction force
        RY = 0.0  # reaction force
        RZ = 0.0  # reaction force
        RXX = -Mx  # reaction moment
        RYY = -My  # reaction moment
        RZZ = -Mz  # reaction moment
        FX = 0.0  # internal force at the support
        FY = 0.0  # internal force at the support
        FZ = 0.0  # internal force at the support
        FXX = Mx  # internal moment at the support
        FYY = My  # internal moment at the support
        FZZ = Mz  # internal moment at the support
        for cell in cells:
            for _ in range(3):
                angles = np.random.rand(3) * np.pi * 2
                u, r, f = solve(n, angles, loads, cell)
                UX_fem, UXX_fem = u[0], u[3]
                RX_fem, RXX_fem = r[0], r[3]
                FX_fem, FXX_fem = f[0], f[3]
                UY_fem, UZZ_fem = u[1], u[5]
                RY_fem, RZZ_fem = r[1], r[5]
                FY_fem, FZZ_fem = f[1], f[5]
                UZ_fem, UYY_fem = u[2], u[4]
                RZ_fem, RYY_fem = r[2], r[4]
                FZ_fem, FYY_fem = f[2], f[4]
                assert self.check_relative_error(UX, UX_fem)
                assert self.check_relative_error(UY, UY_fem)
                assert self.check_relative_error(UZ, UZ_fem)
                assert self.check_relative_error(UXX, UXX_fem)
                assert self.check_relative_error(UYY, UYY_fem)
                assert self.check_relative_error(UZZ, UZZ_fem)
                assert self.check_relative_error(RX, RX_fem)
                assert self.check_relative_error(RY, RY_fem)
                assert self.check_relative_error(RZ, RZ_fem)
                assert self.check_relative_error(RXX, RXX_fem)
                assert self.check_relative_error(RYY, RYY_fem)
                assert self.check_relative_error(RZZ, RZZ_fem)
                assert self.check_relative_error(FX, FX_fem)
                assert self.check_relative_error(FY, FY_fem)
                assert self.check_relative_error(FZ, FZ_fem)
                assert self.check_relative_error(FXX, FXX_fem)
                assert self.check_relative_error(FYY, FYY_fem)
                assert self.check_relative_error(FZZ, FZZ_fem)

    def test_temperature(self):
        """
        Uniform temperature load on a clamped-clamped beam cousing
        normal force only.
        """
        # units in kN, cm and Celsius
        L = 100.0
        Ex, nu = 21000.0, 0.25
        dT = 20
        alpha = 1.2e-5
        # section
        w, h = 5.0, 10.0  # width and height of the rectangular cross section
        A = w * h  # area
        Iy = w * h**3 / 12  # second moment of inertia around the y axis
        Iz = h * w**3 / 12  # second moment of inertia around the z axis
        Ix = Iy + Iz  # torsional inertia
        # material
        G = Ex / (2 * (1 + nu))
        Hooke = np.array(
            [
                [Ex * A, 0, 0, 0],
                [0, G * Ix, 0, 0],
                [0, 0, Ex * Iy, 0],
                [0, 0, 0, Ex * Iz],
            ]
        )

        def build_structure(n: int):
            # space
            GlobalFrame = StandardFrame(dim=3)
            # mesh
            p0 = np.array([0.0, 0.0, 0.0])
            p1 = np.array([L, 0.0, 0.0])
            coords = linspace(p0, p1, n + 1)
            coords = PointCloud(coords, frame=GlobalFrame).show()
            topo = np.zeros((n, 2), dtype=int)
            topo[:, 0] = np.arange(n)
            topo[:, 1] = np.arange(n) + 1
            i_first = index_of_closest_point(coords, np.array([0, 0, 0]))
            i_last = index_of_closest_point(coords, np.array([L, 0, 0]))
            # essential boundary conditions
            fixity = np.zeros((coords.shape[0], 6)).astype(bool)
            fixity[i_first, :] = True
            fixity[i_last, :] = True
            # natural boundary conditions
            nLoadCase = 2
            # nodal loads
            nodal_loads = np.zeros((coords.shape[0], 6, nLoadCase))
            # cell loads
            strain_loads = np.zeros((topo.shape[0], 4, nLoadCase))
            strain_loads[:, 0, 0] = dT * alpha
            strain_loads[:, 0, 1] = -dT * alpha
            # pointdata
            pd = PointData(
                coords=coords, fixity=fixity, loads=nodal_loads, frame=GlobalFrame
            )
            # celldata
            frames = repeat(GlobalFrame.show(), topo.shape[0])
            cd = B2(topo=topo, frames=frames, material=Hooke, strain_loads=strain_loads)
            # set up mesh and structure
            mesh = LineMesh(pd, cd, frame=GlobalFrame)
            structure = Structure(mesh=mesh)
            return structure

        structure = build_structure(2)
        structure.linear_static_analysis()
        forces = structure.internal_forces()
        FX = Ex * A * alpha * dT
        assert self.check_relative_error(-FX, forces[0, 0, 0, 0])
        assert self.check_relative_error(-FX, forces[-1, 0, 0, 0])
        assert self.check_relative_error(FX, forces[0, 0, 0, 1])
        assert self.check_relative_error(FX, forces[-1, 0, 0, 1])

    def test_vs_Navier_LinStat(self):
        """
        Tests all result components against Navier's solution for a
        sinusoidal distributed load. The beam is bent in the x-y plane.
        """
        L, w, h = 1000.0, 20.0, 80.0  # geometry
        Ex, nu = 210000.0, 0.25  # material
        # section properties
        Iy = h * w**3 / 12
        Iz = w * h**3 / 12
        Ix = (Iy + Iz) / 2
        A = w * h
        EI = Ex * Iz
        # material
        G = Ex / (2 * (1 + nu))
        Hooke = np.array(
            [
                [Ex * A, 0, 0, 0],
                [0, G * Ix, 0, 0],
                [0, 0, Ex * Iy, 0],
                [0, 0, 0, Ex * Iz],
            ]
        )
        # load
        qy_max = 1.0  # max intensity
        ny = 1  # number of half waves, should be odd

        def qy(x):
            return np.sin(PI * ny * x / L) * qy_max

        def uy(x):
            return (
                qy_max * (L**4 / (EI * PI**4 * ny**4)) * np.sin(PI * ny * x / L)
            )

        def rotz(x):
            return (
                qy_max * (L**3 / (EI * PI**3 * ny**3)) * np.cos(PI * ny * x / L)
            )

        def kz(x):
            return (
                -qy_max * (L**2 / (EI * PI**2 * ny**2)) * np.sin(PI * ny * x / L)
            )

        def mz(x):
            return -qy_max * (L**2 / (PI**2 * ny**2)) * np.sin(PI * ny * x / L)

        def vy(x):
            return qy_max * (L / (PI * ny)) * np.cos(PI * ny * x / L)

        def solve(n: int, fnc, celltype=B2):
            # space
            GlobalFrame = StandardFrame(dim=3)
            # mesh
            p0 = np.array([0.0, 0.0, 0.0])
            p1 = np.array([L, 0.0, 0.0])
            coords = linspace(p0, p1, n + 1)
            topo = np.zeros((n, 2), dtype=int)
            topo[:, 0] = np.arange(n)
            topo[:, 1] = np.arange(n) + 1
            if celltype.NNODE == 3:
                coords, topo = L2_to_L3(coords, topo)
            # mark some points
            i_first = index_of_closest_point(coords, np.array([0.0, 0.0, 0.0]))
            i_last = index_of_furthest_point(coords, np.array([0.0, 0.0, 0.0]))
            # generate load function
            fnc_loads = fnc(coords[:, 0])
            # essential boundary conditions
            penalty = 1e20  # penalty value for essential BCs
            fixity = np.zeros((coords.shape[0], 6)).astype(bool)
            fixity[i_first, [0, 1, 2, 3, 4]] = True
            fixity[i_last, [1, 2, 3, 4]] = True
            fixity = fixity.astype(float) * penalty
            # natural boundary conditions
            nodal_loads = np.zeros((coords.shape[0], 6))
            cell_loads = np.zeros((topo.shape[0], topo.shape[1], 6))
            # pointdata
            pd = PointData(
                coords=coords, frame=GlobalFrame, loads=nodal_loads, fixity=fixity
            )
            # celldata
            frames = repeat(np.eye(3), topo.shape[0])
            cd = celltype(topo=topo, material=Hooke, frames=frames)
            cell_loads[:, :, 1] = cd.pull(data=fnc_loads)[:, :, 0]
            cd.loads = cell_loads
            # set up mesh and structure
            mesh = LineMesh(pd, cd, frame=GlobalFrame)
            structure = Structure(mesh=mesh)
            structure.linear_static_analysis()
            return structure

        structure = solve(20, qy, B3)
        ui = structure.mesh.cell_dof_solution(
            points=[1 / 4, 3 / 4], rng=[0, 1], target="global", flatten=False
        )
        fi = structure.internal_forces(points=[1 / 4, 3 / 4], rng=[0, 1])
        k = structure.mesh.strains(points=[1 / 4, 3 / 4], rng=[0, 1])
        xi = structure.mesh.cells_coords(points=[1 / 4, 3 / 4], rng=[0, 1])
        u_navier = uy(xi[:, :, 0].flatten())
        u_fem = ui[:, :, 1].flatten()
        self.assertTrue(allclose(u_fem, u_navier, atol=1e-2, rtol=1e-1))
        u_navier = rotz(xi[:, :, 0].flatten())
        u_fem = ui[:, :, -1].flatten()
        self.assertTrue(allclose(u_fem, u_navier, atol=1e-5, rtol=1e-2))
        u_navier = kz(xi[:, :, 0].flatten())
        u_fem = k[:, :, -1].flatten()
        self.assertTrue(allclose(u_fem, u_navier, atol=1e-8, rtol=1e-2))
        u_navier = mz(xi[:, :, 0].flatten())
        u_fem = fi[:, :, -1].flatten()
        self.assertTrue(allclose(u_fem, u_navier, atol=600, rtol=8e-3))
        u_navier = vy(xi[:, :, 0].flatten())
        u_fem = fi[:, :, 1].flatten()
        self.assertTrue(allclose(u_fem, u_navier, atol=8, rtol=0.3))

    def test_Dirichlet(self):
        """
        Tests linear static results when the essential boundary conditions
        are formulated with 'fixity' or by using objects. It is checked if
        the displacement results from the two approaches are the same or not.
        """
        L = 100.0  # length of the console
        w, h = 10.0, 10.0  # width and height of the rectangular cross section
        F = -100.0  # value of the vertical load at the free end
        E = 210000.0  # Young's modulus
        nu = 0.3  # Poisson's ratio
        # cross section
        A = w * h  # area
        Iy = w * h**3 / 12  # second moment of inertia around the y axis
        Iz = h * w**3 / 12  # second moment of inertia around the z axis
        Ix = Iy + Iz  # torsional inertia
        # model stiffness matrix
        G = E / (2 * (1 + nu))
        Hooke = np.array(
            [[E * A, 0, 0, 0], [0, G * Ix, 0, 0], [0, 0, E * Iy, 0], [0, 0, 0, E * Iz]]
        )
        # space
        GlobalFrame = StandardFrame(dim=3)
        # mesh
        nElem = 2  # number of finite elements to use
        p0 = np.array([0.0, 0.0, 0.0])
        p1 = np.array([L, 0.0, 0.0])
        coords = linspace(p0, p1, nElem + 1)
        coords = PointCloud(coords, frame=GlobalFrame).show()
        topo = np.zeros((nElem, 2), dtype=int)
        topo[:, 0] = np.arange(nElem)
        topo[:, 1] = np.arange(nElem) + 1
        # load at the rightmost node
        nodal_loads = np.zeros((coords.shape[0], 6))
        nodal_loads[-1, 2] = F

        # ####################################################
        # DEFINE ESSENTIAL BOUNDARY CONDITIONS THROUGH FIXITY
        # ####################################################

        # support at the leftmost node
        fixity = np.zeros((coords.shape[0], 6)).astype(bool)
        fixity[0, :] = True
        # pointdata
        pd = PointData(
            coords=coords, frame=GlobalFrame, loads=nodal_loads, fixity=fixity
        )
        # celldata
        frames = repeat(GlobalFrame.show(), topo.shape[0])
        cd = B2(topo=topo, material=Hooke, frames=frames)
        # set up mesh and structure
        mesh = LineMesh(pd, cd, frame=GlobalFrame)
        structure = Structure(mesh=mesh)
        # solve
        structure.linear_static_analysis()
        dofsol = structure.nodal_dof_solution()
        umin, umax = minmax(dofsol)

        # ####################################################
        # DEFINE ESSENTIAL BOUNDARY CONDITIONS USING OBJECTS
        # ####################################################

        # support at the leftmost node
        support = NodalSupport(
            x=[0, 0, 0], UX=0.0, UY=0.0, UZ=0.0, UXX=0.0, UYY=0.0, UZZ=0.0
        )
        # pointdata
        pd = PointData(coords=coords, frame=GlobalFrame, loads=nodal_loads)
        # celldata
        frames = repeat(GlobalFrame.show(), topo.shape[0])
        cd = B2(topo=topo, material=Hooke, frames=frames)
        # set up mesh and structure
        mesh = LineMesh(pd, cd, frame=GlobalFrame)
        structure = Structure(mesh=mesh)
        # add the nodal constraints
        structure.constraints.append(support)
        # solve and check
        structure.linear_static_analysis()
        dofsol = structure.nodal_dof_solution()
        _umin, _umax = minmax(dofsol)
        self.assertTrue(allclose(_umin, umin, atol=1e-5))
        self.assertTrue(allclose(_umax, umax, atol=1e-5))

    def test_temperature_inclined(self):
        """
        Simply supported beam with an inclined support at the right end
        and a vertical concentrated force at the middle. The values of
        the normal forces are measured against analytic solution.
        """
        L = 100.0  # length of the console
        E = 210000.0  # Young's modulus
        nu = 0.3  # Poisson's ratio
        F = -100.0  # value of the vertical load at the middle
        # cross section
        w, h = 10.0, 10.0  # width and height of the rectangular cross section
        A = w * h  # area
        Iy = w * h**3 / 12  # second moment of inertia around the y axis
        Iz = h * w**3 / 12  # second moment of inertia around the z axis
        Ix = Iy + Iz  # torsional inertia
        # model stiffness matrix
        G = E / (2 * (1 + nu))
        Hooke = np.array(
            [[E * A, 0, 0, 0], [0, G * Ix, 0, 0], [0, 0, E * Iy, 0], [0, 0, 0, E * Iz]]
        )
        # space
        GlobalFrame = StandardFrame(dim=3)
        # mesh
        nElem = 2  # number of finite elements to use
        p0 = np.array([0.0, 0.0, 0.0])
        p1 = np.array([L, 0.0, 0.0])
        coords = linspace(p0, p1, nElem + 1)
        coords = PointCloud(coords, frame=GlobalFrame).show()
        topo = np.zeros((nElem, 2), dtype=int)
        topo[:, 0] = np.arange(nElem)
        topo[:, 1] = np.arange(nElem) + 1
        # concentrated load at the middle
        x_middle = np.array([L / 2, 0, 0], dtype=float)
        i = index_of_closest_point(coords, x_middle)
        nodal_loads = np.zeros((coords.shape[0], 6))
        nodal_loads[i, 2] = F
        # hinged support at the leftmost node
        support_left = NodalSupport(
            x=[0, 0, 0], UX=0.0, UY=0.0, UZ=0.0, UZZ=0.0, UXX=0.0
        )
        # inclined roller support at the rightmost node
        angle = -30 * PI / 180
        frame = GlobalFrame.rotate_new("Body", [0, angle, 0], "XYZ")
        support_right = NodalSupport(x=[L, 0, 0], frame=frame, UZ=0.0)
        # pointdata
        pd = PointData(coords=coords, frame=GlobalFrame, loads=nodal_loads)
        # celldata
        frames = repeat(GlobalFrame.show(), topo.shape[0])
        cd = B2(topo=topo, material=Hooke, frames=frames)
        # set up mesh and structure
        mesh = LineMesh(pd, cd, frame=GlobalFrame)
        structure = Structure(mesh=mesh)
        # add the nodal constraints
        structure.constraints.append(support_left)
        structure.constraints.append(support_right)
        # solve and check
        structure.linear_static_analysis()
        f = structure.internal_forces(flatten=False)[:, :, 0]
        fmin, fmax = minmax(f)
        self.assertTrue(allclose(fmin, -28.867, atol=1e-3, rtol=None))
        self.assertTrue(allclose(fmax, -28.867, atol=1e-3, rtol=None))


class TestBernoulliNaturalVibrations(unittest.TestCase):
    def test_console(self):
        """
        Simple console with distributed mass. The finite element approximation
        is measured agains a lower (Dunkerley) and an upper bound (Rayleigh)
        estimation. The integrated mass is also verified.
        """
        # all units in kN and cm
        L = 150.0  # length of the console [cm]
        F = 1.0  # value of the vertical load at the free end [kN]
        E = 21000.0  # Young's modulus [kN/cm3]
        nu = 0.3  # Poisson's ratio [-]
        w, h = 5.0, 15.0  # width and height of the rectangular cross section
        g = 9.81  # gravitational acceleration [m/s2]
        density = 7750.0 * 1e-6  # mass density [g/cm3]
        nElem = 20  # number of subdivisons to use
        # calculated data
        A = w * h  # area
        Iy = w * h**3 / 12  # second moment of inertia around the y axis
        Iz = h * w**3 / 12  # second moment of inertia around the z axis
        Ix = Iy + Iz  # torsional inertia
        weight = density * 1e-3 * g  # [kN/cm3]
        dpa = density * A  # density per area [g/cm]
        wpa = weight * A  # weight per area [kN/cm]
        mass = L * dpa  # total mass

        def Dunkerley(N: int = 1):
            mi = mass / N
            h = L / N
            i1 = np.sum(list(range(1, N + 1)))
            i3 = np.sum([i**3 for i in range(1, N + 1)])
            acc_x = i1 * mi * h / (E * A)
            acc_y = i3 * mi * h**3 / (3 * E * Iz)
            acc_z = i3 * mi * h**3 / (3 * E * Iy)
            return np.sqrt(1 / acc_x), np.sqrt(1 / acc_y), np.sqrt(1 / acc_z)

        def Rayleigh(structure: Structure):
            M = structure.mass_matrix(penalize=True)
            u = structure.nodal_dof_solution(flatten=True)
            f = structure.nodal_forces(flatten=True)
            return np.sqrt(Rayleigh_quotient(M, u=u, f=f))

        # model stiffness matrix
        G = E / (2 * (1 + nu))
        Hooke = np.array(
            [[E * A, 0, 0, 0], [0, G * Ix, 0, 0], [0, 0, E * Iy, 0], [0, 0, 0, E * Iz]]
        )
        # space
        GlobalFrame = StandardFrame(dim=3)
        # mesh
        p0 = np.array([0.0, 0.0, 0.0])
        p1 = np.array([L, 0.0, 0.0])
        coords = linspace(p0, p1, nElem + 1)
        coords = PointCloud(coords, frame=GlobalFrame).show()
        topo = np.zeros((nElem, 2), dtype=int)
        topo[:, 0] = np.arange(nElem)
        topo[:, 1] = np.arange(nElem) + 1
        # load at the rightmost node in Y and Z directions
        nodal_loads = np.zeros((coords.shape[0], 6, 3))
        global_load_vector = Vector([F, 0, 0.0], frame=GlobalFrame).show()
        nodal_loads[-1, :3, 0] = global_load_vector
        global_load_vector = Vector([0.0, F, 0.0], frame=GlobalFrame).show()
        nodal_loads[-1, :3, 1] = global_load_vector
        global_load_vector = Vector([0.0, 0, F], frame=GlobalFrame).show()
        nodal_loads[-1, :3, 2] = global_load_vector
        # support at the leftmost node (all degrees)
        fixity = np.zeros((coords.shape[0], 6)).astype(bool)
        fixity[0, :] = True
        fixity = fixity.astype(float) * 1e40
        # mass and density
        nodal_masses = np.zeros((coords.shape[0],))
        densities = np.full((topo.shape[0],), dpa)
        # pointdata
        pd = PointData(
            coords=coords, loads=nodal_loads, fixity=fixity, mass=nodal_masses
        )
        # celldata
        frames = repeat(GlobalFrame.show(), topo.shape[0])
        cd = B2(topo=topo, frames=frames, material=Hooke, density=densities)
        # set up mesh and structure
        mesh = LineMesh(pd, cd, frame=GlobalFrame)
        structure = Structure(mesh=mesh)
        # perform linear analysis
        structure.linear_static_analysis()
        # calculate natural vibrations
        structure.cleanup()
        structure.free_vibration_analysis(normalize=True, as_dense=True)
        f, v = structure.natural_circular_frequencies(return_vectors=True)
        # calculate effective masses
        nN, nD = len(coords), 6
        action_x = np.zeros((nN, nD))
        action_x[:, 0] = 1.0
        action_x = action_x.reshape(nN * nD)
        action_y = np.zeros((nN, nD))
        action_y[:, 1] = 1.0
        action_y = action_y.reshape(nN * nD)
        action_z = np.zeros((nN, nD))
        action_z[:, 2] = 1.0
        action_z = action_z.reshape(nN * nD)
        actions = np.stack([action_x, action_y, action_z], axis=1)
        M = structure.mass_matrix(penalize=False)
        m_eff = effective_modal_masses(M, actions, v)
        mass_fem = np.sum(m_eff[:, 0]), np.sum(m_eff[:, 1]), np.sum(m_eff[:, 2])
        mass_percentages = np.array([100 * m / mass for m in mass_fem])
        self.assertTrue(np.all(mass_percentages > 98.0))
        # check values
        imax = [np.argmax(m_eff[:, i]) for i in range(3)]
        f_lower = Dunkerley(20)
        f_upper = Rayleigh(structure)
        f_fem = f[imax]
        c = [f_fem[i] > f_lower[i] and f_fem[i] < f_upper[i] for i in range(3)]
        self.assertTrue(np.all(c))
        # check mass
        mass_fem = structure.mesh.mass()
        self.assertTrue(allclose(mass, mass_fem, atol=1e-3, rtol=None))

    def test_beam_ss(self, celltype=B3):
        """
        Simply supported beam with a hinge at the midspan. The essential
        boundary conditions are specified with using both 'fixity' and
        objects. Mass calculation is checked.
        """
        # all units in kN and cm
        L = 150.0  # length of the console [cm]
        F = 1.0  # value of the vertical load at the free end [kN]
        E = 21000.0  # Young's modulus [kN/cm3]
        nu = 0.3  # Poisson's ratio [-]
        w, h = 5.0, 15.0  # width and height of the rectangular cross section
        g = 9.81  # gravitational acceleration [m/s2]
        density = 7750.0 * 1e-6  # mass density [g/cm3]
        nElem = 4  # number of subdivisons to use
        # calculated data
        A = w * h  # area
        Iy = w * h**3 / 12  # second moment of inertia around the y axis
        Iz = h * w**3 / 12  # second moment of inertia around the z axis
        Ix = Iy + Iz  # torsional inertia
        weight = density * 1e-3 * g  # [kN/cm3]
        dpa = density * A  # density per area [g/cm]
        wpa = weight * A  # weight per area [kN/cm]
        mass = L * dpa  # total mass
        # model stiffness matrix
        G = E / (2 * (1 + nu))
        Hooke = np.array(
            [[E * A, 0, 0, 0], [0, G * Ix, 0, 0], [0, 0, E * Iy, 0], [0, 0, 0, E * Iz]]
        )
        # space
        GlobalFrame = StandardFrame(dim=3)
        # mesh
        p0 = np.array([0.0, 0.0, 0.0])
        p1 = np.array([L, 0.0, 0.0])
        coords = linspace(p0, p1, nElem + 1)
        coords = PointCloud(coords, frame=GlobalFrame).show()
        topo = np.zeros((nElem, 2), dtype=int)
        topo[:, 0] = np.arange(nElem)
        topo[:, 1] = np.arange(nElem) + 1
        if celltype.NNODE == 3:
            coords, topo = L2_to_L3(coords, topo)
        # load at the rightmost node in Y and Z directions
        nodal_loads = np.zeros((coords.shape[0], 6, 3))
        global_load_vector = Vector([F, 0, 0.0], frame=GlobalFrame).show()
        nodal_loads[-1, :3, 0] = global_load_vector
        global_load_vector = Vector([0.0, F, 0.0], frame=GlobalFrame).show()
        nodal_loads[-1, :3, 1] = global_load_vector
        global_load_vector = Vector([0.0, 0, F], frame=GlobalFrame).show()
        nodal_loads[-1, :3, 2] = global_load_vector
        # support at the leftmost node (all degrees)
        fixity = np.zeros((coords.shape[0], 6)).astype(bool)
        fixity[0, :] = True
        fixity = fixity.astype(float) * 1e12
        # support at the rightmost node
        support = NodalSupport(
            x=[L, 0, 0], UX=0.0, UY=0.0, UZ=0.0, UXX=0.0, UYY=0.0, UZZ=0.0
        )
        # mass and density
        nodal_masses = np.zeros((coords.shape[0],))
        densities = np.full((topo.shape[0],), dpa)
        # pointdata
        pd = PointData(
            coords=coords, loads=nodal_loads, fixity=fixity, mass=nodal_masses
        )
        # cell fixity / end releases
        cell_fixity = np.full((topo.shape[0], topo.shape[1], 6), True)
        cell_fixity[int(nElem / 2 - 1), -1, 4] = False
        # celldata
        frames = repeat(GlobalFrame.show(), topo.shape[0])
        cd = celltype(
            topo=topo,
            frames=frames,
            density=densities,
            fixity=cell_fixity,
            material=Hooke,
        )
        # set up mesh and structure
        mesh = LineMesh(pd, cd, frame=GlobalFrame)
        structure = Structure(mesh=mesh, essential_penalty=1e12, mass_penalty_ratio=1e5)
        structure.constraints.append(support)
        # perform vinration analysis
        structure.free_vibration_analysis(normalize=True, as_dense=True)
        structure.natural_circular_frequencies(return_vectors=True)
        # check mass
        mass_fem = structure.mesh.mass()
        self.assertTrue(allclose(mass, mass_fem, atol=1e-3, rtol=None))


if __name__ == "__main__":
    unittest.main()
